# -*- coding: utf-8 -*-
"""
"""

from . import __version__
from .log import logger


class ExitCode:
    EXPECTED_EXIT = 0

    NO_CONFIG_FOUND = 1
    INVALID_CONFIG = 2
    DUPLICATE_CONFIG_SECTIONS = 3
    DUPLICATE_CONFIG_OPTIONS = 4
    CONFIG_WRITE_ERROR = 5
    NO_CURRENT_VERSION = 6
    ROLLBACK_NOT_POSSIBLE = 7

    GIT_UNUSABLE = 10
    GIT_ROOT_ERROR = 11
    GIT_DESCRIBE_ERROR = 12

    VERSION_PARSE_ERROR = 20
    VERSION_METADATA_ERROR = 21
    VERSION_BUMP_ERROR = 22
    VERSION_PEP440_ERROR = 23
    VERSION_GIT_ERROR = 24

    SOURCE_CODE_FILE_OPEN_ERROR = 30
    SOURCE_CODE_FILE_MISSING = 31
    USER_SUPPLIED_REGEX_ERROR = 32

    PIPELINE_NOT_FOUND = 40
    INVALID_INTERNAL_CMD = 42


class IcanException(Exception):
    def __init__(self, *args, **kwargs):
        """ """
        self.output_method = logger.critical
        self.exit_code = self.__class__.exit_code
        self.msg = ""
        self.e = []

        # Set the message, typically at init
        if args:
            if isinstance(args[0], Exception):
                self.msg = str(args[0])

                _e = args[0]
                if hasattr(_e, "stderr"):
                    stderr = getattr(_e, "stderr")
                    stderr = stderr.decode("utf-8")
                    stderr = stderr.rstrip("\n")
                    self.e.append(stderr)
                if hasattr(_e, "returncode"):
                    returncode = getattr(_e, "returncode")
                    self.e.append(returncode)
            else:
                self.msg = args[0]
        elif hasattr(self.__class__, "msg"):
            self.msg = self.__class__.msg

        # Merge the exit_code, msg, and version
        m = f"{self.msg} (code-{self.exit_code}) v{__version__}"
        self.msg = m

    def __str__(self):
        return self.msg


class DryRunExit(IcanException):
    pass


########################


class NoConfigFound(IcanException):
    exit_code = ExitCode.NO_CONFIG_FOUND
    msg = "Cannot find ican config.  Maybe try `ican init`"


class InvalidConfig(IcanException):
    exit_code = ExitCode.INVALID_CONFIG


class DuplicateConfigSections(IcanException):
    exit_code = ExitCode.DUPLICATE_CONFIG_SECTIONS
    msg = "Duplicate config sections. Look for duplicate file/pipeline labels."


class DuplicateConfigOptions(IcanException):
    exit_code = ExitCode.DUPLICATE_CONFIG_OPTIONS
    msg = "Duplicate config options.  Look for duplicate steps in a pipeline."


class ConfigWriteError(IcanException):
    exit_code = ExitCode.CONFIG_WRITE_ERROR


class NoCurrentVersion(IcanException):
    exit_code = ExitCode.NO_CURRENT_VERSION
    msg = (
        "[NO_VERSION_SPECIFIED]\n"
        "Check if current version is specified in config file, like:\n"
        "version = 0.4.3\n"
    )


class RollbackNotPossible(IcanException):
    exit_code = ExitCode.ROLLBACK_NOT_POSSIBLE
    msg = "Rollback is not possible because no previous version exists."


########################


class GitUnusable(IcanException):
    exit_code = ExitCode.GIT_UNUSABLE


class GitRootError(IcanException):
    exit_code = ExitCode.GIT_ROOT_ERROR


class GitDescribeError(IcanException):
    exit_code = ExitCode.GIT_DESCRIBE_ERROR


########################


class VersionParseError(IcanException):
    exit_code = ExitCode.VERSION_PARSE_ERROR


class VersionMetadataError(IcanException):
    exit_code = ExitCode.VERSION_METADATA_ERROR


class VersionBumpError(IcanException):
    exit_code = ExitCode.VERSION_BUMP_ERROR


class VersionPep440Error(IcanException):
    exit_code = ExitCode.VERSION_PEP440_ERROR


class VersionGitError(IcanException):
    exit_code = ExitCode.VERSION_GIT_ERROR


########################


class SourceCodeFileOpenError(IcanException):
    exit_code = ExitCode.SOURCE_CODE_FILE_OPEN_ERROR


class SourceCodeFileMissing(IcanException):
    exit_code = ExitCode.SOURCE_CODE_FILE_MISSING


class UserSuppliedRegexError(IcanException):
    exit_code = ExitCode.USER_SUPPLIED_REGEX_ERROR


########################


class PipelineNotFound(IcanException):
    exit_code = ExitCode.PIPELINE_NOT_FOUND


class InvalidInternalCmd(IcanException):
    exit_code = ExitCode.INVALID_INTERNAL_CMD




# -*- coding: utf-8 -*-
"""
"""


class Registry(object):
    def __init__(self):
        self.reg = {}

    def set(self, name, instance=None):
        if instance is None:
            raise ValueError("Instance cannot be None")
        self.reg[name] = instance

    def get(self, name):
        return self.reg.get(name)


registry = Registry()


class Base:
    def __init__(self):
        pass

    @property
    def version(self):
        return registry.get("version")

    @version.setter
    def version(self, value):
        registry.set("version", value)

    @property
    def config(self):
        return registry.get("config")

    @config.setter
    def config(self, value):
        registry.set("config", value)

    @property
    def git(self):
        return registry.get("git")

    @git.setter
    def git(self, value):
        registry.set("git", value)

    @property
    def ican(self):
        return registry.get("ican")

    @ican.setter
    def ican(self, value):
        registry.set("ican", value)



# -*- coding: utf-8 -*-
"""
from tempfile import NamedTemporaryFile
"""

import argparse
import sys

from . import __version__
from .base import Base
from .config import Config
from .ican import Ican
from .log import logger
from .exceptions import IcanException
from .exceptions import PipelineNotFound


# ===============================
#
#  CLI Class
#
# ===============================


class CLI(Base):
    usage = """ican <COMMAND> [<ARGS>]

commands:
  bump [PART]        increment version [minor, major, patch, prerelease, build]
  pre [TOKEN]        set the prerelease string [alpha, beta, rc, dev]
  show [STYLE]       show version [semantic, public, pep440, git]
  run [PIPELINE]     run the specified PIPELINE
  rollback           restore the previous version
  init               initialize a config in the current directory
"""

    def __init__(self):
        """
        """
        self._register_excepthook()
        # self._arg_pop()
        self.config = Config()

        parser = argparse.ArgumentParser(usage=CLI.usage, prog="ican")
        parser.add_argument("command", help=argparse.SUPPRESS)
        parser.add_argument(
            "--version",
            action="version",
            help="display ican version",
            version=f"ican v{__version__}",
        )

        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            # Unknown, look for pipelines with this name (alias)
            try:
                self.run(args.command)
            except PipelineNotFound:
                logger.error("Unrecognized command")
                parser.print_help()
                exit(1)
            except Exception as e:
                raise(e)

        else:
            getattr(self, args.command)()
            return

    def _register_excepthook(self):
        """Register our custom exception handler"""

        self._original_excepthook = sys.excepthook
        sys.excepthook = self._excepthook
        return

    def _excepthook(self, type, value, tracekback, debug=False):
        """Custom exception handler"""

        if isinstance(value, IcanException):
            if value.msg:
                value.output_method(value.msg)
            if value.e:
                for line in value.e:
                    value.output_method(line)
            if debug:
                self._original_excepthook(type, value, tracekback)
            exit_code = value.exit_code
            sys.exit(exit_code)
        else:
            self._original_excepthook(type, value, tracekback)

    def _arg_pop(self):
        """Here we will pop --verbose and --dry_run out asap,
        that way logging can be setup before we parse the config
        file, etc.
        """

        verbose = False
        dry_run = False
        for i in range(1, len(sys.argv)):
            if sys.argv[i] == "--verbose":
                verbose = True
                sys.argv.pop(i)
                break
        for i in range(1, len(sys.argv)):
            if sys.argv[i] == "--dry_run":
                dry_run = True
                sys.argv.pop(i)
                break
        logger.setup(verbose, dry_run)
        return

    def command_prep(self, parser, alias=None, user_args=False):
        # self._arg_pop()
        self.config.pre_parse()

        parser.add_argument(
            "--dry_run", help="do not write any files", action="store_true"
        )
        parser.add_argument(
            "--verbose", help="display debug information", action="store_true"
        )

        argv = sys.argv[2:]
        if alias:
            argv.insert(0, alias)

        args = parser.parse_args(argv)
        logger.setup(args.verbose, args.dry_run)
        return args

    def bump(self):
        """dispatched here with command bump"""

        parser = argparse.ArgumentParser(
            description="PART choices [major, minor, patch, [pre]release, build]",
            usage="ican bump [PART]",
        )
        parser.add_argument(
            "part",
            nargs="?",
            default="build",
            choices=["major", "minor", "patch", "prerelease", "pre", "build"],
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            "--pre",
            type=str,
            help="set the prerelease string [alpha, beta, rc, dev]",
        )
        args = self.command_prep(parser)

        self.ican = Ican()
        self.ican.bump(args.part.lower(), args.pre)
        logger.verbose("bump() COMPLETE")
        return

    def show(self):
        """dispatched here with command show"""

        parser = argparse.ArgumentParser(
            description="STYLE choices [semantic, public, pep440, git]",
            usage="ican show [STYLE]",
        )
        parser.add_argument(
            "style",
            nargs="?",
            default="semantic",
            choices=["semantic", "public", "pep440", "git"],
            help=argparse.SUPPRESS,
        )
        args = self.command_prep(parser)

        self.ican = Ican(only_pre_parse=True)
        self.ican.show(args.style)
        return

    def pre(self):
        """Sets the prerelease token to [alpha, beta, dev, rc]
        At the same time prerelease mode is enabled for the
        version.  The same can be accomplished with --pre
        during a bump.
        """

        parser = argparse.ArgumentParser(
            description="set the prerelease TOKEN to one of "
            "[alpha, beta, rc, dev]",
            usage="ican pre [TOKEN]",
        )
        parser.add_argument(
            "token",
            nargs="?",
            default="beta",
            choices=["alpha", "beta", "rc", "dev"],
            help=argparse.SUPPRESS,
        )
        args = self.command_prep(parser)

        self.ican = Ican(only_pre_parse=True)
        self.ican.pre(args.token)
        return

    def rollback(self):
        """in case of emergency, restore the previously
        persisted version.
        """

        parser = argparse.ArgumentParser(usage="ican rollback")
        self.command_prep(parser)

        self.ican = Ican()
        self.ican.rollback(only_pre_parse=True)
        logger.verbose("rollback() COMPLETE")
        return

    def init(self):
        """dispatched here with command init"""

        parser = argparse.ArgumentParser(usage="ican init")
        self.command_prep(parser)

        self.config = Config(init=True).parse()
        logger.info("init COMPLETE")

        return

    def run(self, alias=None):
        """dispatched here with command init"""

        parser = argparse.ArgumentParser(
            description="PIPELINE can be any pipeline defined in your .ican file.",
            usage="ican run [PIPELINE]",
        )
        parser.add_argument("pipeline", help=argparse.SUPPRESS)
        parser.add_argument(
            'user_args',
            nargs="*",
            help=argparse.SUPPRESS
        )
        args = self.command_prep(parser, alias)

        self.ican = Ican()
        self.ican.run_pipeline(args.pipeline, args.user_args)
        logger.info(f"+FINISHED pipeline.{args.pipeline.upper()}")

    def test2(self):
        """dispatched here with command test"""

        parser = argparse.ArgumentParser(usage="ican test [ARGS]")
        # parser.add_argument("dynamic_args", nargs="*", help=argparse.SUPPRESS)
        parser.add_argument(
            'user_args',
            nargs=argparse.REMAINDER,
            help=argparse.SUPPRESS)
        args = self.command_prep(parser)

        logger.verbose("verbose")
        print(f"10-4 with arg {args.dynamic_args}")


def entry():
    CLI()



# -*- coding: utf-8 -*-

from .base import Base
from .version import Version
from .git import Git
from .log import logger
from .exceptions import GitDescribeError
from .exceptions import NoConfigFound
from .exceptions import RollbackNotPossible
from .exceptions import PipelineNotFound


#######################################
#
#   Bump Class
#
#######################################


class Ican(Base):
    """
    Object which will orchestrate entire program
    """

    def __init__(self, only_pre_parse=False):
        """Typically ican will be instantiated by cli with a half parsed
        config.  We pre-parse so logging can begin.
        """
        self.ready = False

        # make sure the config is fully parsed
        if not self.config.pre_parsed:
            # Typically 'init' is the ONLY way to be in this state
            self.config.parse()
        elif not self.config.parsed and not only_pre_parse:
            self.config.parse()
        # Here if still config not ready, it will never be ready
        if not self.config.config_file:
            raise NoConfigFound()

        # Now config is parsed.  We can parse from config
        self.version = Version.parse(self.config.current_version)
        logger.verbose(f"discovered {self.version.semantic} @ CONFIG.version")

        # Git init
        self.git = Git()

        try:
            self.version._git_metadata = self.git.describe()
        except GitDescribeError as e:
            logger.verbose(e)
            logger.verbose("Git-versions are disabled. Does this repo have a tag?")
            self.git.disable()
        else:
            logger.verbose(f"Discovered {self.version.git} @ GIT.version")
        return

    def pre(self, token):
        """Set the prerelease token"""

        logger.verbose(f"Setting prerelease string to {token}")
        self.version.set_prerelease_token(token)

        # Save the new version, config will check for dry_run
        self.config.persist_version(self.version.semantic)

        # Display message to user
        logger.info(f"prerelease token set: {self.version.semantic}")
        return self

    def show(self, style="semantic"):
        """
        Show the <STYLE> version
        """

        v = None
        if hasattr(self.version, style):
            v = getattr(self.version, style)
        if v is None:
            logger.error(f"Version.Style not available: {style}")
        # Display message to user
        logger.info(f"Current {style} version: {v}")
        return self

    def rollback(self):
        """When all else fails, this should bring the version back
        to your prior saved version.  It will also update all source
        files you have configured.
        """
        if not self.config.previous_version:
            raise RollbackNotPossible()

        # delete old, create new self.version
        del self.version
        self.version = Version.parse(self.config.previous_version)

        # Update the source files
        for file in self.config.source_files:
            file.update(self.version)

        # Now that everything else is finished, persist version
        self.config.persist_version(self.config.previous_version)

        # Display message to user
        logger.info(f"Rollback: {self.version.semantic}")
        return self

    def bump(self, part="build", token=None):
        """This is pretty much the full process"""
        if token:
            # Little magic here to turn `bump pre` into `bump prerelease`
            if token == "pre":
                token = "prerelease"
            logger.verbose(f"Setting prerelease string to {token}")
        logger.verbose(f"Beginning bump of <{part.upper()}>")

        self.version.bump(part, token)
        logger.verbose(
            f"New value of <{part.upper()}> - {getattr(self.version, part)}"
        )

        # Update the user's files with new version
        for file in self.config.source_files:
            file.update(self.version)

        # Once all else is successful, persist the new version
        self.config.persist_version(self.version.semantic)

        # Display message to user
        logger.info(f"Version: {self.ican.version.semantic}")
        return self

    def run_pipeline(self, pipeline, user_args=[]):
        # Pipeline
        if self.config.pipelines.get(pipeline) is None:
            # Pipeline is not defined
            raise PipelineNotFound(f'pipeline.{pipeline.upper()} not found')

        pl = self.config.pipelines.get(pipeline)
        pl.run(user_args)
        return self



# -*- coding: utf-8 -*-

import logging
import sys

# debug=10, info=20, warning=30, error=40, critical=50


class IcanFormatter(logging.Formatter):
    """colors from https://stackoverflow.com/a/56944256/3638629"""

    BLACK = "\u001b[30;1m"
    RED = "\u001b[31;1m"
    GREEN = "\u001b[32;1m"
    YELLOW = "\u001b[33;1m"
    BLUE = "\u001b[34;1m"
    MAGENTA = "\u001b[35;1m"
    CYAN = "\u001b[36;1m"
    WHITE = "\u001b[37;1m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    BLINK = "\033[5m"
    INVERT = "\033[7m"
    RESET = "\u001b[0m"

    DEFAULT_FORMAT = "%(filename)-12s  %(message)s"

    def __init__(self, fmt=DEFAULT_FORMAT, emoji=True):
        super().__init__()
        self.fmt = fmt
        self.plain = "%(message)s"
        self.FORMATS = {}
        self.FORMATS[logging.DEBUG] = self.color(self.MAGENTA)
        self.FORMATS[IcanLogger.VERBOSE] = self.color(self.YELLOW)
        self.FORMATS[IcanLogger.DRY_RUN] = self.color(self.GREEN)
        self.FORMATS[IcanLogger.ALT_INFO] = self.color(self.INVERT + self.CYAN, True)
        self.FORMATS[logging.INFO] = self.color(self.UNDERLINE + self.BLUE, True)
        self.FORMATS[logging.WARNING] = self.color(self.RED)
        self.FORMATS[logging.ERROR] = self.color(self.RED)
        self.FORMATS[logging.CRITICAL] = self.color(self.INVERT + self.RED)

        if emoji:
            pass

    def color(self, format, plain=False):
        if plain:
            return format + self.plain + self.RESET
        return format + self.fmt + self.RESET

    def format(self, record):
        """Use the FORMATS dict to apply color, also we strip
        .py from the record.filename.
        """

        record.filename = record.filename.replace(".py", "").upper()
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class IcanLogger(logging.getLoggerClass()):
    """
    assert logging.getLevelName(VERBOSE) == 'VERBOSE'
    assert logging.getLevelName(DRY_RUN) == 'DRY_RUN'
    """

    VERBOSE = 11  # like debug, print on console and file
    DRY_RUN = 15  # dry_run related messages
    ALT_INFO = 19   # really info but different color to stand out in run_pipeline
    DEFAULT_FILEHANDLER = "%(asctime)s | %(levelname)s | %(message)s"

    def __init__(self, name, level=logging.DEBUG):
        super().__init__(name, level)
        self._verbose = None
        self._dry_run = None
        logging.addLevelName(self.VERBOSE, "VERBOSE")
        logging.addLevelName(self.DRY_RUN, "DRY_RUN")
        logging.addLevelName(self.ALT_INFO, "ALT_INFO")
        self.ready = False

    def _welcome_msg(self):
        self.info("        ---===::: Welcome to ican :::===---")
        self.verbose("--verbose detected.  Displaying verbose messaging.")
        self.dry_run("--dry_run detected.  No files will be modified.")

    def _set_dry_run(self, val=True):
        self._dry_run = val

    def _set_verbose(self, val=True):
        self._verbose = val

    def setup(self, verbose=False, dry_run=False):
        """We call this right away to configure logging.  We will store both
        verbose and dry_run here so we can log verbosely without worry and
        we can use the logger object to determine file writes with the
        property ok_to_write
        """

        console = logging.StreamHandler(sys.stderr)
        if dry_run:
            self._set_dry_run()
            console.setLevel(self.DRY_RUN)
        if verbose:
            self._set_verbose()
            console.setLevel(self.VERBOSE)
        if not verbose and not dry_run:
            console.setLevel(logging.INFO)

        console.setFormatter(IcanFormatter())
        self.addHandler(console)
        self._welcome_msg()
        self.ready = True

    def setup_file_handler(self, filename, format=DEFAULT_FILEHANDLER):
        """This method sets up our filehandler.  This is separate as console
        logging can begin right away but we have to parse the config to get
        the filename for file based logs.
        """

        formatter = logging.Formatter(format, "%m-%d-%Y %H:%M:%S")
        file_handler = logging.FileHandler(filename)
        # File handler gets all messages for now
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        self.addHandler(file_handler)

    def verbose(self, msg, *args, **kwargs):
        """Custom logging method for our VERBOSE logging level.  Nothing
        special it is just like the stock DEBUG, INFO, etc.
        """

        if self.isEnabledFor(self.VERBOSE):
            self._log(self.VERBOSE, msg, args, **kwargs)

    def dry_run(self, msg, *args, **kwargs):
        """This is a bit unusual.  Dry_run is 15, so someone with
        VERBOSE, 11, would typically get these msgs as well.  But,
        we ONLY want dry_run people to see them, so we have an
        extra if statement built in.
        """

        if self.isEnabledFor(self.DRY_RUN):
            if self._dry_run:
                self._log(self.DRY_RUN, msg, args, **kwargs)

    def alt_info(self, msg, *args, **kwargs):
        """Custom logging method for our VERBOSE logging level.  Nothing
        special it is just like the stock DEBUG, INFO, etc.
        """

        if self.isEnabledFor(self.ALT_INFO):
            self._log(self.ALT_INFO, msg, args, **kwargs)

    @property
    def ok_to_write(self):
        """Every module already imports the logger so this was a convenient
        place to globally determine if we can write via the dry_run flag.
        Also we auto-log a msg when this returns false.
        """

        if self._dry_run:
            self.dry_run("Detected --dry_run.  File modification denied.")
            return False
        return True


logging.setLoggerClass(IcanLogger)
logger = logging.getLogger("ican")



# -*- coding: utf-8 -*-

#
#   ,---.  ,-.,---.  ,---.  ,-.    ,-..-. .-.,---.
#   | .-.\ |(|| .-.\ | .-'  | |    |(||  \| || .-'
#   | |-' )(_)| |-' )| `-.  | |    (_)|   | || `-.
#   | |--' | || |--' | .-'  | |    | || |\  || .-'
#   | |    | || |    |  `--.| `--. | || | |)||  `--.
#   /(     `-'/(     /( __.'|( __.'`-'/(  (_)/( __.'
#  (__)      (__)   (__)    (_)      (__)   (__)
#

import os
import re
import subprocess
import shlex
from collections import UserDict

from .log import logger
from .base import Base

from .exceptions import InvalidInternalCmd


#######################################
#
#   Step
#
#######################################


class CTX(UserDict):
    def __missing__(self, key):
        if key.startswith("arg_"):
            return ""
        return "#NOT_FOUND#"

    def __setitem__(self, item, value):
        # Make sure it's a string
        value = str(value)
        super().__setitem__(item, value)

    def gen_env(self):
        env = {f'ICAN_{k.upper()}': v for k, v in self.items()}
        return {**os.environ, **env}


class Step:

    COMMANDS = r"\$ICAN\(\s*?(?P<cmd>.*?)\s*?\)"
    CLI = "__CLI__"
    INTERNAL = "__INTERNAL__"

    def __init__(self, label, cmd):
        self.label = label
        self.cmd = None
        self.backup = None
        self.type = None

        match = re.search(self.COMMANDS, cmd)
        if match:
            self.cmd = match.group("cmd")
            self.type = self.INTERNAL
        else:
            self.cmd = cmd
            self.type = self.CLI

    def is_internal(self):
        if self.type == self.INTERNAL:
            return True
        return False

    def is_cli(self):
        if self.type == self.CLI:
            return True
        return False

    def render(self, ctx):
        """render template variables
        """

        formatted = self.cmd.format_map(ctx)
        if formatted != self.cmd:
            self.backup = self.cmd
            self.cmd = formatted
            logger.verbose(f"rendered cmd: {formatted}")
        return


#######################################
#
#   Pipeline
#
#######################################


class Pipeline(Base):

    ARGS = r"{arg_(?P<num>\d+)}"
    CLI = "__CLI__"
    INTERNAL = "__INTERNAL__"
    INTERNAL_CMDS = {
        'bump': 'bump',
        'run': 'run_pipeline',
        'pre': 'pre',
        'rollback': 'rollback',
        'show': 'show',
    }

    def __init__(self, label=None, steps=None):
        self.label = label
        self.steps = []
        self.env = None
        self.ctx = None

        if steps is None:
            logger.error("must include at least 1 step")

        for k, v in steps:
            logger.verbose(f"{label.upper()}.{k} - {v}")
            step = Step(k, v)
            self.steps.append(step)

    def _run_cli_cmd(self, step):
        """Here is where we actually run the cli commands.
        Args:
            cmd: This should be a tuple or list of command, args such as:
            ['git', 'commit', '-a']

        Returns:
            result: the result object will have attributes of both
            stdout and stderr representing the results of the subprocess
        """
        if not logger.ok_to_write:
            return

        if type(step.cmd) not in (tuple, list):
            step.cmd = shlex.split(step.cmd)
        logger.verbose(f"running cmd - {step.cmd}")
        result = subprocess.run(
            step.cmd,
            shell=False,
            env=self.env,
            capture_output=False,
            text=True
        ).stdout

        if result:
            logger.verbose(f"cmd result - {result}")
        return

    def _run_internal(self, step):
        """This is for pipelines with itnernal commands.  Such as:
        $_ICAN{bump build} runs version.bump('build')
        """
        if not logger.ok_to_write:
            return

        logger.verbose(f"running internal cmd - {step.cmd}")
        parts = step.cmd.split(' ')
        cmd = parts[0].lower()
        if cmd not in self.INTERNAL_CMDS.keys():
            raise InvalidInternalCmd()

        args = [x for x in parts[1:] if x]
        # Run using same dispatch method as cli
        getattr(self.ican, self.INTERNAL_CMDS.get(cmd))(*args)

        return

    def _build_ctx(self, user_args):
        """ """

        ctx = CTX()
        ctx["version"] = self.version.semantic
        ctx["semantic"] = self.version.semantic
        ctx["public"] = self.version.public
        ctx["pep440"] = self.version.pep440
        ctx["git"] = self.version.git
        ctx["tag"] = self.version.tag
        ctx["major"] = self.version.major
        ctx["minor"] = self.version.minor
        ctx["patch"] = self.version.patch
        ctx["prerelease"] = self.version.prerelease
        ctx["build"] = self.version.build
        ctx["env"] = self.version.env
        ctx["age"] = self.version.age
        ctx["root"] = self.config.path
        ctx["previous"] = self.config.previous_version

        # include the user_args
        x = 0
        for arg in user_args:
            x += 1
            ctx[f"arg_{x}"] = arg

        logger.verbose(f"Generated ctx: {ctx}")
        # Use the ctx to generate a new env
        self.env = ctx.gen_env()
        return ctx

    def run(self, user_args):
        print(user_args)
        logger.alt_info(f"+BEGIN pipeline.{self.label.upper()}")
        for step in self.steps:
            # Rebuild the ctx each step
            ctx = self._build_ctx(user_args)

            # Each step renders their template variables
            step.render(ctx)
            logger.alt_info(F"+RUNNING step.{step.type.upper()}<{step.cmd}>")
            if step.is_internal():
                self._run_internal(step)
            if step.is_cli():
                self._run_cli_cmd(step)

        logger.alt_info(f"+END pipeline.{self.label.upper()}")



# -*- coding: utf-8 -*-

import subprocess
import os

from pathlib import Path
from types import SimpleNamespace

from .log import logger
from .exceptions import GitDescribeError
from .exceptions import GitRootError

#######################################
#
#   Git Class
#
#######################################


"""
commit message keywords borrowed from python-semantic-release

feat: A new feature. (minor)
fix: A bug fix. (patch)
docs: Documentation changes. (build)
style: Changes that do not affect the meaning of the code
    (white-space, formatting, missing semi-colons, etc). (build)
refactor: A code change that neither fixes a bug nor adds a feature. (build)
perf: A code change that improves performance. (patch)
test: Changes to the test framework. (build)
build: Changes to the build process or tools. (build)
"""


class Git(object):
    """
    Simple class to communicate with git via subprocess git commands.
    The other alternative is to use python-git which is big and rarely
    already installed.
    """

    def __init__(self):
        # First we'll initialize some variables
        self.usable = None
        self.root = None

        # Now we can test git, get root_dir, and describe
        self.usable = self.is_usable()
        if self.usable:
            self.root = self.find_root()
            r = str(self.root).rstrip("\n")
            logger.verbose(f"found git root @ {r}")

    def disable(self):
        self.usable = False

    def command(self, cmd=[]):
        env = os.environ.copy()

        result = subprocess.run(cmd, check=True, capture_output=True, env=env).stdout

        return result

    def is_usable(self):
        """
        Simple test to check if git is installed and can
        be used.
        """

        cmd = ["git", "rev-parse", "--git-dir"]
        try:
            u = self.command(cmd)
        except Exception:
            return False

        return u

    def find_root(self):
        """
        Find the git root
        Returns:
            Returns: pathlib.Path instance of the root git dir if found.
        """

        cmd = ["git", "rev-parse", "--show-toplevel"]
        try:
            dir = self.command(cmd)
        except Exception as e:
            raise GitRootError(e)

        return Path(dir.decode())

    def describe(self):
        """
        get info about the latest tag in git
        """

        cmd = ["git", "describe", "--dirty", "--tags", "--long"]
        try:
            d = self.command(cmd).decode().split("-")
        except Exception as e:
            raise GitDescribeError(e)

        dirty = False
        if d[-1].strip() == "dirty":
            dirty = True
            d.pop()

        commit_sha = d.pop().lstrip("g")
        distance = int(d.pop())
        tag = "-".join(d).lstrip("v")

        g = SimpleNamespace(
            tag=tag,
            commit_sha=commit_sha,
            distance=distance,
            dirty=dirty,
        )
        return g



# -*- coding: utf-8 -*-
"""

   ___ ___  _  _ ___ ___ ___
  / __/ _ \| \| | __|_ _/ __|
 | (_| (_) | .` | _| | | (_ |
  \___\___/|_|\_|_| |___\___|.py

"""

import os
from pathlib import Path
from configparser import ConfigParser
from configparser import DuplicateSectionError
from configparser import DuplicateOptionError
from configparser import ParsingError
from collections import OrderedDict

from .base import Base
from .source import SourceCode
from .pipeline import Pipeline
from .log import logger
from .exceptions import DuplicateConfigSections
from .exceptions import DuplicateConfigOptions
from .exceptions import ConfigWriteError
from .exceptions import InvalidConfig


#######################################
#
#   Config Class
#
#######################################


class MultiOrderedDict(OrderedDict):
    """Custom obj to use as parser storage to allow multiple
    keys with same keyname. Credit:
    https://stackoverflow.com/questions/15848674/
    """
    def __setitem__(self, key, value):
        if isinstance(value, list) and key in self:
            self[key].extend(value)
        else:
            # super(MultiOrderedDict, self).__setitem__(key, value)
            super().__setitem__(key, value)


class Config(Base):
    """Config object for entire program
    """

    default_ver = dict(current="0.1.0")
    default_options = dict(log_file="ican.log")
    default_file = dict(file="*.py", style="semantic", variable="__version__")

    CONFIG_FILE = ".ican"
    DEFAULT_CONFIG = dict(
        version=default_ver, options=default_options, file1=default_file
    )

    def __init__(self, init=False):
        self.config_file = None
        self.ran_from = Path.cwd()
        self.parser = ConfigParser()

        self.current_version = None
        self.previous_version = None
        self.log_file = None
        self.source_files = []
        self.pipelines = {}
        self.pre_parsed = False
        self.parsed = False

        if init:
            self.init()
        return

    @property
    def path(self):
        if self.config_file:
            return self.config_file.parent.absolute()
        return None

    def ch_dir_root(self):
        """chdir to the config root, so if we run from another dir, relative
        file paths, etc still work as expected.
        """
        if self.path:
            os.chdir(str(self.path).rstrip("\n"))
        return

    def save(self):
        if logger.ok_to_write:
            if not self.config_file:
                f = Path(self.ran_from, Config.CONFIG_FILE)
                self.config_file = f
            try:
                self.parser.write(open(self.config_file, "w"))
                logger.verbose("wrote config file")
            except Exception as e:
                raise ConfigWriteError(e)
        return

    def persist_version(self, new_version):
        """Update the version in the config file then write it so we
        know the new version next time.  Save previous in case
        we need it to rollback.
        """
        logger.verbose(f"persisting version - {new_version}")

        self.previous_version = self.current_version
        self.parser.set("version", "previous", self.previous_version)

        self.current_version = new_version
        self.parser.set("version", "current", self.current_version)

        self.save()
        return

    def init(self):
        """Set default config and save"""
        logger.verbose("cmd: init - setting default config")
        self.parser.read_dict(Config.DEFAULT_CONFIG)
        self.save()
        return self

    def locate_config_file(self):
        """Find our config file."""
        logger.verbose("Searching for config file")
        f = Config.CONFIG_FILE
        dir = Path.cwd()
        root = Path(dir.root)
        while True:
            cfg = Path(dir, f)
            if cfg.exists():
                self.config_file = cfg
                logger.verbose(f"Config found @ {cfg}")
                return True
            dir = dir.parent
            if dir == root:
                # quit right before we'd be writing in the root
                logger.verbose("Cannot find config file!")
                break
        return None

    def pre_parse(self):
        """Partially parse the config.  Enough to log and grab the version.
        """

        if not self.config_file:
            # Only time we could already have self.config_file is if we
            # ran an init
            if not self.locate_config_file():
                # Silently continue on.  There is still a chance for init()
                return self
            try:
                self.parser.read([self.config_file])
            except DuplicateSectionError as e:
                raise DuplicateConfigSections(e)
            except DuplicateOptionError as e:
                raise DuplicateConfigOptions(e)
            except ParsingError as e:
                raise InvalidConfig(e)

        self.current_version = self.parser.get(
            "version",
            "current",
            fallback="0.1.0"
        )
        self.previous_version = self.parser.get(
            "version",
            "previous",
            fallback=[None]
        )
        self.log_file = self.parser.get(
            "options",
            "log_file",
            fallback=[None]
        )

        self.ch_dir_root()
        if self.log_file:
            logger.setup_file_handler(self.log_file)
        self.pre_parsed = True
        return self

    def parse(self):
        """The parse() method parses the entire config file.  You
        can pre_parse before running parse or not.  Either way it should
        all end up parsed.
        """
        if not self.pre_parsed:
            self.pre_parse()

        self.parse_source_files()
        self.parse_pipelines()
        self.parsed = True
        return self

    def parse_pipelines(self):
        for s in self.parser.sections():
            if not s.startswith("pipeline:"):
                # Not interested in this section
                continue

            label = s.split(":")[1].strip().lower()
            logger.verbose(f"parsing {label.upper()} pipeline")
            left_right_tuple = self.parser.items(s)
            pl = Pipeline(label=label, steps=left_right_tuple)
            self.pipelines[label] = pl
        return

    def parse_source_files(self):
        # FILES TO WRITE
        for s in self.parser.sections():
            if not s.startswith("file:"):
                # Not interested in this section
                continue

            label = s.split(":")[1].strip().lower()
            file = self.parser.get(s, "file", fallback=None)
            style = self.parser.get(s, "style", fallback="semantic")
            variable = self.parser.get(s, "variable", fallback=None)
            regex = self.parser.get(s, "regex", fallback=None)

            # Instead of raising exp, we can just look for more files
            if file is None:
                logger.verbose(f"skipping source - missing file ({label})")
                continue
            elif variable is None and regex is None:
                logger.verbose("Skipping source - missing variable/regex")
                continue

            logger.verbose(f"parsing file config {label.upper()}[{file}]")
            # Case with *.py for all python files
            if "*" in file:
                files = self._find_wildcard_filename(file)
            else:
                files = [file]
            for f in files:
                u = SourceCode(label, f, style=style, variable=variable, regex=regex)
                self.source_files.append(u)

    def _find_wildcard_filename(self, f):
        """Search for all files if config has '*' in the
        filename field.  Search root dir + all subdirs.
        """

        logger.verbose(f"file section * in filename - {f}")
        matches = [x for x in Path(self.path).rglob(f)]
        if matches:
            logger.verbose(f"wildcard found: {len(matches)} files")
            return matches
        return None
